import 'package:petitparser/petitparser.dart';

import 'env.dart';
import 'nodes.dart';
import 'runtime.dart';
import 'utils.dart';

class Parser {
  Parser(this.env) {
    _comma.set(wtsp & char(',') & wtsp);
    _colon.set(wtsp & char(':') & wtsp);
    _lBracket.set(char('[') & wtsp);
    _rBracket.set(wtsp & char(']'));
    _lCurly.set(char('{') & wtsp);
    _rCurly.set(wtsp & char('}'));
    _int.set(_hex
            .flatten()
            .map((values) => Literal<int>(int.tryParse(values[1], radix: 16))) |
        _floor.flatten().map((str) => Literal<int>(int.tryParse(str))));
    _prec.set(char('.') & _floor);
    _exp.set(pattern('eE') & char('-').optional() & _floor);
    _double.set((_floor & _prec & _exp.optional())
        .flatten()
        .map((str) => Literal<double>(double.tryParse(str))));
    _string.set(((char('"') & stringContetDQ.star() & char('"')) |
            (char('\'') & stringContentSQ.star() & char('\'')))
        .flatten()
        .map((str) => Literal<String>(str)));
    _literals.set(_string | _double | _int | _false | _true | _none);
    _listLiteral.set(
        (_lBracket.flatten() & _exprList & _rBracket.flatten()).map((values) {
      final values_ = values[1];
      if (values_ == null) return const Literal<Null>(null);
      if (values_.isEmpty) return const Literal<List>([]);
      if (values_.every((value) => value is Literal))
        return Literal<List>(
            values_.map((value) => value.value).toList(growable: false));
      return ExpressionList(values_);
    }));
    __mapEntry.set((_string.flatten() & _colon.flatten() & _primary)
        .map((values) => MapEntry(values[0], values[2])));
    _mapLiteral.set((_lCurly.flatten() &
            (__mapEntry & _comma.optional()).map((values) => values[0]).star() &
            _rCurly.flatten())
        .map((values) {
      values = values[1];
      if (values.every((value) => value.value is Literal))
        return Map.fromEntries(
            values.map((value) => MapEntry(value.key, value.value.value)));
      return ExpressionMap.of(
          Map.fromEntries(values.cast<MapEntry<String, Expression>>()));
    }));
    _primary.set(_mapLiteral | _listLiteral | _literals | _identifier);

    _expr.set(_primary.end());
    _exprList
        .set((_primary & _comma.optional()).map((values) => values[0]).star());
  }

  final Environment env;

  final _exprList = undefined();
  final _expr = undefined();
  final _primary = undefined();
  final _mapLiteral = undefined();
  final __mapEntry = undefined();
  final _listLiteral = undefined();
  final _literals = undefined();
  final _string = undefined();
  final stringContetDQ = pattern('^\\"\n\r') | char('\\') & pattern('\n\r');
  final stringContentSQ = pattern("^\\'\n\r") | char('\\') & pattern('\n\r');
  final _double = undefined();
  final _int = undefined();
  final _hex = string('0x').seq(pattern('0-9a-fA-F').plus());
  final _exp = undefined();
  final _prec = undefined();
  final _floor = digit().plus();
  final _false = string('false').map((_) => Literal<bool>(false));
  final _true = string('true').map((_) => Literal<bool>(true));
  final _none = string('none').map((_) => null);
  final _identifier = letter().plus().flatten().map((str) => Variable(str));
  final _lCurly = undefined();
  final _rCurly = undefined();
  final _lBracket = undefined();
  final _rBracket = undefined();
  final _colon = undefined();
  final _comma = undefined();
  final wtsp = pattern(' \t').plus().optional();
  final nwln = pattern('\n\r');

  parse(String source, [path]) => _expr.parse(source).value;
}
